import GenericRobot
import RobotCanvas
import RobotFrame
import Tkinter
import time
import threading
import random
import Queue
import math
import sys


class ThreadHandler:
    """
    Launches the main part of the GUI and the worker thread. periodicCall and
    endApplication could reside in the GUI part, but putting them here
    means that you have all the thread controls in a single place.
    """
    def __init__(self, mapname):
        """
        Start the GUI and the asynchronous threads. We are in the main
        (original) thread of the application, which will later be used by
        the GUI. We spawn a new thread for the worker.
        """
        # the map
        self.mapname = mapname
        
        # basic windows
        self.root = Tkinter.Tk()
        self.root.withdraw()
        self.master = Tkinter.Toplevel(master=self.root)
        
        # Create the queue for message passing with the graphics
        self.queue = Queue.Queue()
        
        # Set up the GUI part
        self.rf = RobotFrame.RobotFrame(self.master)
        ##self.gui = VisualizerThread.VisualizerThread(self.master, self.queue, self.mapname)
        self.canv = self.rf.canv  # for easy reference
        # map-loading before creating robot sprites...
        self.canv.loadMap(self.mapname)
        time.sleep(0.05)

        # create the graphical object representing the robot's odometric position
        self.odrobot = RobotCanvas.MovableCircle( self.rf.canv, colorstr='blue' )
        self.simrobot = RobotCanvas.MovableCircle( self.rf.canv, colorstr='red' )
        
        # the robot simulator, r and the map it is in...
        self.r = GenericRobot.GenericRobot(0,self.mapname)
        time.sleep(1)

        # create two more graphical objects (start them off-screen)
        self.rangeLine = RobotCanvas.MovableLine(self.canv, 
            [(-10000,-10000),(-10000,-10000)], colorstr='red', width=1)
        self.goalPoint = RobotCanvas.MovablePoint(self.canv, cx = -10000,
             cy = -10000, pixelradius=5, colorstr='green')

        # see if this helps with getting the graphics set up
        time.sleep(1)
        
        # Set up a thread to run the robot's main loop (main)
        # more threads can be made, as needed
        self.sim = threading.Thread(target=self.main)
        self.sim.setDaemon(1)  # it will quit when the GUI does
        self.sim.start()       # starts the main loop thread

        # the particles and their forward-looking ranges
        self.ranges = []
        self.particles = []
        
        # Starts the GUI call to check if the queue contains anything
        self.checkMessages()

    def createParticles(self, N):
        """ creates N particles for display """
        P = len(self.particles)
        if N < P:
            for i in range(N,P):
                self.particles[i].setGlobalPose(-424242.0, -424242.0, 0.0)
                self.ranges[i].changeCoords([(0,0),(0,0)])
           
        if N > P:
            for i in range(P,N):
                self.particles.append( RobotCanvas.MovableCircle( self.rf.canv, colorstr='yellow', radius=4 ) )
                self.particles[-1].setGlobalPose(-424242.0, -424242.0, 0.0)
                self.ranges.append( RobotCanvas.MovableLine(self.rf.canv, 
                      [(-10000,-10000),(-10000,-10000)], colorstr='red', width=1) )
        # that's it, maybe...?

    def setParticles(self, locations):
        """ sets the particles' locations to those held in locations """
        P = len(self.particles);
        L = len(locations)
        if P != L:
            self.createParticles(L)
            
        for i in range(L):
            self.particles[i].setGlobalPose( locations[i][0], locations[i][1], locations[i][2] )
            self.ranges[i].setGlobalPose(locations[i][0], locations[i][1], locations[i][2])
            col = min([255, locations[i][3]*128])
            self.particles[i].setColor(0,col,0)
            self.ranges[i].changeCoords([(0,0),(locations[i][4],0)])
        
    def run(self):
        """ starts handling graphical events """
        try:
            self.root.mainloop()
        except KeyboardInterrupt:  # user hit control-c at the command line
            print 'Control-c caught from GUI. Quitting...'
            sys.exit(0)

    def getRealData(self):
        data = None
        return data
        
    def getWindowEvents(self):
        return self.canv.lastKeyDown, self.canv.lastMouseClick
        
    def getRealPose(self):
        return self.r.gpos.x, self.r.gpos.y, math.degrees(self.r.gpos.thr)

    def shutdown(self):
        """ shuts down the graphics
            since the main loop is a daemon thread, it will automatically quit
        """
        self.rf.quit()          # quit the graphics
        time.sleep(0.5)
        #self.master.destroy()   # quit Tkinter's root object
        sys.exit(0)             # leave python entirely
        
    def getData(self):
        """ get all of the sensor data of interest to the robot
        """
        # get the pose
        (x,y,thr), (odx,ody,odthr), bump = self.r.getUpdatedPose()
        # update the graphics to reflect this data
        self.queue.put( (self.simrobot.setGlobalPose, x, y, thr) )
        self.queue.put( (self.odrobot.setGlobalPose, odx, ody, odthr) )
        
        # get the range/camera reading
        # this variable allows one to point the range sensor in any direction
        # for now, it's fixed, pointing forward
        RANGEFINDERHEADINGDEG = 0.0
        dist, rd, gr, bl = self.r.getColorAndRange(RANGEFINDERHEADINGDEG)
        odx, ody, odthr = self.r.odpos.vals()
        # update the graphics to reflect this data
        if dist > 0:
            self.queue.put( (self.rangeLine.setGlobalPose,odx,ody,odthr+math.radians(RANGEFINDERHEADINGDEG)) )
            self.queue.put( (self.rangeLine.setColor,rd,gr,bl) )
            self.queue.put( (self.rangeLine.changeCoords, ((0,0),(dist,0)) ) )
        else:
            # if no reading, do not draw (rather, draw a zero-length 
            dist = 0.0
            self.queue.put( (self.rangeLine.changeCoords, ((0,0),(dist,0)) ) )

        # return realpose, odometric pose, bump status, range sensor distance, and color
        return ( (x,y,math.degrees(thr)), (odx,ody,odthr), bump, dist, (rd, gr, bl) )


    def getRangeData( self, x, y, thr, rangeSensorHeadingDeg=0.0 ):
        """ returns a four-tuple of dist, rd, gr, bl
        """
        return self.r.getColorAndRange( rangeSensorHeadingDeg, (x,y,thr) )
        
            
        
    def setVels(self, cmpersec, degpersec, noise=0.0):
        """ velocity-setting
                first it checks limits, then it includes noise
        """
        if cmpersec >  50.0: cmpersec =  50.0
        if cmpersec < -50.0: cmpersec = -50.0
        if degpersec >  75.0: degpersec =  75.0
        if degpersec < -75.0: degpersec = -75.0

        # delegate to the robot
        self.r.setVels(cmpersec, degpersec, noise)

    def getMap(self):
        """ convenience accessor """
        return self.r.simmap.items
        
    def showGoal(self, goal):
        """ send to graphics side """
        self.queue.put( (self.goalPoint.setGlobalPose, goal[0], goal[1], 0.0) )

    def checkMessages(self):
        """
        Handle all the messages currently in the queue (if any).
        """
        # redraw the graphics
        self.canv.redrawRemote()

        while self.queue.qsize():  # while the queue has messages
            try:
                # L is a list representing a command to execute
                L = self.queue.get(0)
                # L[0] is the name of the method to run
                # L[1:] are its parameters
                L[0](*L[1:])
            except Queue.Empty:    # if empty, the while loop will break
                pass

        # reschedule this for 20 milliseconds later (50 hz)
        self.master.after(20, self.checkMessages)
        
                
    def main(self):
        """ placeholder for the MainLoop's main method
        """
        # we should not be here
        print 'In MainClientThread\'s main method'
        
        
        
